#!/usr/bin/env python3
# XPathGenerator.pl
# Alpino-XML XPath Generator

# version 2.0 data: 10.11.2021 translated to Python, multiple includes/excludes per token
# version 1.7 date: 10.06.2015  bug fix (@number)
# version 1.6 date: 15.12.2014  bug fix (ignore not-function if word order is checked)
# version 1.5 date: 14.10.2014  RELEASED WITH GrETEL2.0
# written by Vincent Vandeghinste and Liesbeth Augustinus (c) 2014
# for the GrETEL2.0 project

# script converts an XML tree into an XPath expression

############################################################################
# argument: -xml: path to xml-subtree
# options: -order/-o: word order is important
#          -r: exclude root node from the XPath expression
#          -version/-v: script details
#          -in: attributes to include (comma-separated list)
#          -ex: attributes to exclude (comma-separated list)

############################################################################

import re
from lxml import etree


def generate_xpath(twig: etree._Element, order: bool) -> str:
    root = twig

    if root.xpath('/alpino_ds'):
        # for ALPINO XML, leave out the alpino_ds node
        subtree = root.find('node')
    else:
        subtree = root    # start at root node

    # generate XPath expression

    topxpath, negate = GetXPath(subtree)
    xpath = ProcessTree(subtree, order)

    if xpath and topxpath:    # if more than one node is selected
        if negate:
            xpath = f'//not({topxpath} and {xpath}])'
        else:
            xpath = f'//{topxpath} and {xpath}]'

    elif xpath and not topxpath:
        xpath = f'//*[{xpath}]'

    elif not xpath and topxpath:
        # if only one node is selected
        if negate:
            xpath = f'//not({topxpath}])'
        else:
            xpath = f'//{topxpath}]'

    else:
        print("ERROR: no XPath expression could be generated.\n")

    return xpath


def ProcessTree(tree, order):
    xpath = ''
    children = tree.getchildren()
    childxpaths = []
    COUNTS = {}
    ALREADY = set()
    if len(children) > 0:
        for child in children:
            childxpath, negate = GetXPath(child)

            if childxpath:
                lower = ProcessTree(child, order)
                if lower:
                    childxpath += f' and {lower}]'

                else:
                    childxpath += ']'

                if negate:
                    childxpath = f"not({childxpath})"
                COUNTS[childxpath] = COUNTS.get(childxpath, 0) + 1
                childxpaths.append(childxpath)

        if childxpaths:
            i = 0
            while (i < len(childxpaths)):

                # ADD COUNT FUNCTION
                if COUNTS[childxpaths[i]] > 1:
                    childxpaths[i] = \
                        'count(' \
                        + childxpaths[i] + ') > ' \
                        + str(COUNTS[childxpaths[i]] - 1)

                # REMOVE DOUBLE DAUGHTERS
                if childxpaths[i] in ALREADY:
                    childxpaths = childxpaths[:i] + childxpaths[i+1:]
                    i -= 1

                else:
                    ALREADY.add(childxpaths[i])

                i += 1

            xpath = str.join(' and ', childxpaths)

        else:
            # die "not implemented yet\n";
            return None

    else:    # no children
        if order:
            xpath = 'number(@begin)'
            next_term, nextpath = FindNextTerminalToCompare(tree)
            if next_term is not None:
                if float(tree.attrib.get('begin', 'nan')) < float(next_term.attrib.get('begin', 'nan')):

                    xpath += " < "

                else:
                    xpath += " > "

                xpath += nextpath

            else:
                return None

    return xpath


def FindNextTerminalToCompare(tree):
    next_sibling = tree.getnext()
    if next_sibling is not None:
        path = "../"
        next_terminal, xpath = FindNextLeafNode(next_sibling)
        path = path + xpath
        if 'begin' in path:

            # $path='number('.$path.')';
            path = re.sub(r'\@begin', 'number(@begin)', path)

    else:
        # go up the tree to find next sibling
        parent = tree.getparent()
        if parent is not None:
            next_terminal, nextpath = FindNextTerminalToCompare(parent)
            if not nextpath:
                return None, None

            path = "../" + nextpath

        else:
            return None, None

    return next_terminal, path


def FindNextLeafNode(node):
    children = node.getchildren()
    xpath, negate = GetXPath(node)
    if negate:
        xpath = f"not({xpath}])"
    else:
        xpath += "]"
    if len(children) > 0:
        node, childpath = FindNextLeafNode(children[0])
        xpath += "/" + childpath
        return node, xpath

    else:
        path = xpath + '/@begin'
        return node, path


def property_selector(key: str, value: str, lower: bool, negative: bool) -> str:
    operator = "!=" if negative else "="

    if lower:
        selector = f"lower-case(@{key}){operator}\"{value.lower()}\""
    elif value:
        selector = f"@{key}{operator}\"{value}\""
    else:
        selector = f"@{key}"
        if negative:
            selector = f"not({selector})"

    return selector


def GetXPath(tree):
    att = tree.attrib
    exclude = att.get('exclude', '').split(',')

    selectors = []
    # if all selectors are exclusive, use the positive selectors
    # and negate the entire node
    positive_selectors = []

    for key in att:
        # all attributes are included in the XPath expression...
        # ...except these ones
        if key not in ['postag', 'begin', 'end', 'caseinsensitive', 'exclude']:
            value = att[key]
            lower = False

            if value and key in ['word', 'lemma']:
                caseinsensitive = att.get('caseinsensitive', 'no')
                if caseinsensitive == 'yes':
                    lower = True

            if key in exclude:
                selectors.append(property_selector(key, value, lower, True))
                positive_selectors.append(
                    property_selector(key, value, lower, False))
            else:
                selectors.append(property_selector(key, value, lower, False))

    if not selectors:

        # no matching attributes found
        return '', False

    else:
        # one or more attributes found
        if len(selectors) == len(positive_selectors):
            selectors = positive_selectors
            negate = True
        else:
            negate = False
        string = str.join(" and ", selectors)
        xstring = "node[" + string

        return xstring, negate


def main(inputxml, order):
    twig = etree.fromstring(bytes(inputxml, encoding='utf-8'))
    if order in ['false', 'False', '0', 0, False]:
        order = False
    else:
        order = True

    xpath = generate_xpath(twig, order)
    return xpath



#!/usr/bin/env python3

# GetSubtree.pl
# Subtree finder which extracts a subtree from an Alpino XML tree

# version 2.0 data: 10.11.2021 translated to Python, multiple includes/excludes per token
# version 1.8 date: 23.05.2018 made more robust for new versions of Alpino
# version 1.7 date: 15.12.2014  bug fix
# version 1.6 date: 15.10.2014  RELEASED WITH GrETEL2.0
# written by Liesbeth Augustinus (c) 2014
# for the GrETEL2.0 project

#########################################################################

# argument: -xml: path to xml-tree (with @include and @exclude attributes for the nodes that should be kept)
# options:
#     - rel: remove top rel
#     - cat: remove top cat
#     - relcat: remove top rel and top cat

#########################################################################

import re
from typing import Optional, List, Union
from lxml import etree


def generate_subtree(twig: etree._Element, remove: Union[str, List[str]]) -> Optional[etree._Element]:
    refpos = initialize()

    # start at 'top' node (leave out alpino_ds node, skip 'parser' tag)
    # 1.8 added 'node'restriction to make sure the first_child is the syntax tree
    root = twig.find('node')  # start at 'top' node (leave out alpino_ds node)

    subtree = process_twig(root, refpos)
    if subtree is None:
        # nothing to match, so match everything
        return None
    top = cut_unary(subtree)

    # Remove top node attrib, except when it's the only node
    if remove and len(top.getchildren()):
        if 'rel' in remove and 'rel' in top.attrib:
            del top.attrib['rel']
        if 'cat' in remove and 'cat' in top.attrib:
            top.attrib['cat'] = ""

    return top


""" subroutines """


def cut_unary(twig):
    children = twig.getchildren()
    if len(children) == 1:
        return cut_unary(children[0])
    else:
        return twig


def process_twig(twig, refpos):
    children = twig.getchildren()
    for child in children:
        result = process_twig(child, refpos)
        if result is None:
            twig.remove(child)
        else:
            # preserve @cat and @rel if there are children
            update = str.join(',', set(twig.attrib.get(
                'include', '').split(',') + ['cat', 'rel']))
            twig.attrib['include'] = update

    if ('include' in twig.attrib or 'exclude' in twig.attrib) \
            and twig.attrib.get('include', '') != 'na':
        include = twig.attrib.get('include', '').split(',')
        exclude = twig.attrib.get('exclude', '').split(',')
        hash = twig.attrib

        preserve = include + exclude

        # needed for generating queries preserving the word order
        preserve.append('begin')
        preserve.append('exclude')

        if 'cs' not in include:
            preserve.append('caseinsensitive')

        for key in hash:
            if key not in preserve:
                del twig.attrib[key]

        if 'postag' in twig.attrib:
            cgntag = twig.attrib['postag']    # get CGN postag
            # split tag into separate attribute-value pairs
            split = split_one_tag(cgntag, refpos)
            if split:
                for s in split:
                    [att, val] = s.split('|')
                    twig.attrib[att] = val    # add new elements
                    if 'postag' in exclude:
                        exclude.append(att)
            twig.attrib['exclude'] = ','.join(exclude)

        return twig

    else:
        return None


def split_one_tag(tag, refpos):

    # split tag
    # refpos = reference naar hash
    match = re.match(r'(\w+)\((.*?)\)', tag)
    if match:
        pt = match.group(1)                # get pt
        pts = match.group(2)                # get other parts

    # assign attribute to parts
    pts = pts.split(',')    # split parts
    parts = []
    for val in pts:

        if pt != 'BW' or pt != 'TSW' or pt != 'LET':
            feature = refpos[pt]
            att = feature[val]    # same as $att=$feature->{$val};

        else:
            # do nothing if $pt equals BW, TSW or LET
            return None

        attval = att + '|' + val       # combine attribute-value
        parts.append(attval)

    return parts    # return array of attribute-value pairs


def initialize():

    # hashes with value-attribute pairs

    n = {
        'soort': 'ntype',
        'eigen': 'ntype',
        'ev': 'getal',
        'mv': 'getal',
        'basis': 'graad',
        'dim': 'graad',
        'onz': 'genus',
        'zijd': 'genus',
        'stan': 'naamval',
        'gen': 'naamval',
        'dat': 'naamval'
    }

    adj = {
        'prenom': 'positie',
        'nom': 'positie',
        'post': 'positie',
        'vrij': 'positie',
        'basis': 'graad',
        'comp': 'graad',
        'sup': 'graad',
        'dim': 'graad',
        'zonder': 'buiging',
        'met-e': 'buiging',
        'met-s': 'buiging',
        'zonder-n': 'getal-n',
        'mv-n': 'getal-n',
        'stan': 'naamval',
        'bijz': 'naamval'
    }

    ww = {
        'pv': 'wvorm',
        'inf': 'wvorm',
        'od': 'wvorm',
        'vd': 'wvorm',
        'tgw': 'pvtijd',
        'verl': 'pvtijd',
        'conj': 'pvtijd',
        'ev': 'pvagr',
        'mv': 'pvagr',
        'met-t': 'pvagr',
        'prenom': 'positie',
        'nom': 'positie',
        'vrij': 'positie',
        'zonder': 'buiging',
        'met-e': 'buiging',
        'zonder-n': 'getal-n',
        'mv-n': 'getal-n'
    }

    tw = {
        'hoofd': 'numtype',
        'rang': 'numtype',
        'prenom': 'positie',
        'nom': 'positie',
        'vrij': 'positie',
        'zonder-n': 'getal-n',
        'mv-n': 'getal-n',
        'basis': 'graad',
        'dim': 'graad',
        'stan': 'naamval',
        'bijz': 'naamval'
    }

    vnw = {
        'pers': 'vwtype',
        'refl': 'vwtype',
        'pr': 'vwtype',
        'recip': 'vwtype',
        'pos': 'vwtype',
        'vrag': 'vwtype',
        'betr': 'vwtype',
        'bez': 'vwtype',
        'vb': 'vwtype',
        'excl': 'vwtype',
        'aanw': 'vwtype',
        'onbep': 'vwtype',
        'pron': 'pdtype',
        'adv-pron': 'pdtype',
        'det': 'pdtype',
        'grad': 'pdtype',
        'stan': 'naamval',
        'nomin': 'naamval',
        'obl': 'naamval',
        'gen': 'naamval',
        'dat': 'naamval',
        'vol': 'status',
        'red': 'status',
        'nadr': 'status',
        '1': 'persoon',
        '2': 'persoon',
        '2v': 'persoon',
        '2b': 'persoon',
        '3': 'persoon',
        '3p': 'persoon',
        '3m': 'persoon',
        '3v': 'persoon',
        '3o': 'persoon',
        'ev': 'getal',
        'mv': 'getal',
        'getal': 'getal',
        'masc': 'genus',
        'fem': 'genus',
        'onz': 'genus',
        'prenom': 'positie',
        'nom': 'positie',
        'post': 'positie',
        'vrij': 'positie',
        'zonder': 'buiging',
        'met-e': 'buiging',
        'met-s': 'buiging',
        'agr': 'npagr',
        'evon': 'npagr',
        'rest': 'npagr',
        'evz': 'npagr',
        'agr3': 'npagr',
        'evmo': 'npagr',
        'rest3': 'npagr',
        'evf': 'npagr',

        # 'mv'=> 'npagr',
        'zonder-n': 'getal-n',
        'mv-n': 'getal-n',
        'basis': 'graad',
        'comp': 'graad',
        'sup': 'graad',
        'dim': 'graad'
    }

    lid = {
        'bep': 'lwtype',
        'onbep': 'lwtype',
        'stan': 'naamval',
        'gen': 'naamval',
        'dat': 'naamval',
        'agr': 'npagr',
        'evon': 'npagr',
        'evmo': 'npagr',
        'rest': 'npagr',
        'rest3': 'npagr',
        'evf': 'npagr',
        'mv': 'npagr'
    }

    vz = {
        'init': 'vztype',
        'versm': 'vztype',
        'fin': 'vztype'
    }

    vg = {
        'neven': 'vgtype',
        'onder': 'vgtype'
    }

    spec = {
        'afgebr': 'spectype',
        'onverst': 'spectype',
        'vreemd': 'spectype',
        'deeleigen': 'spectype',
        'meta': 'spectype',
        'comment': 'spectype',
        'achter': 'spectype',
        'afk': 'spectype',
        'symb': 'spectype'
    }

    # hash of hash references
    refpos = {}
    refpos['N'] = n
    refpos['ADJ'] = adj
    refpos['WW'] = ww
    refpos['TW'] = tw
    refpos['VNW'] = vnw
    refpos['LID'] = lid
    refpos['VZ'] = vz
    refpos['VG'] = vg
    refpos['SPEC'] = spec

    return refpos    # {} => hash reference


def main(inputxml: str, remove: str) -> Optional[etree._Element]:
    twig = etree.fromstring(bytes(inputxml, encoding='utf-8'))
    subtree = generate_subtree(twig, remove)
    return subtree



#!/usr/bin/env python3
from typing import List, Optional
from lxml import etree

from .marker import main as mark
from .subtree import generate_subtree
from .xpath_generator import main as generate_xpath


class AlpinoQuery:
    subtree: Optional[etree._Element]

    @property
    def marked_xml(self) -> str:
        return self.__get_xml(self.marked)

    @marked_xml.setter
    def marked_xml(self, value: str) -> None:
        self.marked = etree.fromstring(bytes(value, encoding='utf-8'))

    @property
    def subtree_xml(self):
        if self.subtree is None:
            return '<node cat="top" rel="top"></node>\n'
        return self.__get_xml(self.subtree)

    @subtree_xml.setter
    def subtree_xml(self, value: str) -> None:
        self.subtree = etree.fromstring(bytes(value, encoding='utf-8'))

    def mark(self, inputxml: str, tokens: List[str], attributes: List[str]) -> etree._Element:
        self.marked = mark(inputxml, tokens, attributes)
        return self.marked

    def generate_subtree(self, remove: List[str]) -> None:
        """
        Generate subtree, removes the top "rel" and/or "cat"
        """
        self.subtree = generate_subtree(self.marked, remove)

    def generate_xpath(self, order: bool) -> str:
        self.xpath = generate_xpath(self.subtree_xml, order)
        return self.xpath

    def __get_xml(self, twig) -> str:
        return etree.tostring(twig, pretty_print=True).decode()



#!/usr/bin/env python3
# mark the matched tokens in the tree with include/exclude attributes
import re
from typing import cast, List, Set
from lxml import etree


def mark(twig: etree._Element, tokens: List[str], attributes: List[str]) -> None:
    # add info annotation matrix to alpino parse
    for begin, token in enumerate(tokens):
        if re.match(r"([_<>\.,\?!\(\)\"\'])|(\&quot;)|(\&apos;)", token):
            xp = cast(List[etree._Element], twig.xpath(f"//node[@begin='{begin}']"))
        else:
            xp = cast(List[etree._Element], twig.xpath(f"//node[@word='{token}' and @begin='{begin}']"))
        if begin < len(attributes):
            attrs = attributes[begin].split(',')
            for x in xp:
                include: Set[str] = set()
                exclude: Set[str] = set()
                case_insensitive = None
                for attr in attrs:
                    if attr[0] == '-':
                        target = exclude
                        attr = attr[1:]
                    else:
                        target = include

                    if attr == 'cs':
                        # cs: case sensitive
                        # -cs: case insensitive
                        case_insensitive = target is exclude
                    elif attr == 'word':
                        target.add('word')
                        if case_insensitive is None:
                            case_insensitive = True
                    else:
                        target.add(attr)

                if case_insensitive:
                    x.attrib['caseinsensitive'] = 'yes'

                if include:
                    x.attrib['include'] = str.join(',', sorted(include))
                if exclude:
                    x.attrib['exclude'] = str.join(',', sorted(exclude))


def main(inputxml: str, tokens: List[str], attributes: List[str]) -> etree._Element:
    twig: etree._Element = etree.fromstring(bytes(inputxml, encoding='utf-8'))
    mark(twig, tokens, attributes)
    return twig



from setuptools import setup, find_packages

with open('README.md') as file:
    long_description = file.read()

setup(
    name='alpino-query',
    python_requires='>=3.6, <4',
    version='2.1.6',
    description='Generating XPATH queries based on a Dutch Alpino syntax tree and user-specified token properties.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Digital Humanities Lab, Utrecht University',
    author_email='digitalhumanities@uu.nl',
    url='https://github.com/UUDigitalHumanitieslab/alpino-query',
    license='CC BY-NC-SA 4.0',
    packages=['alpino_query'],
    zip_safe=True,
    install_requires=[
        'lxml'
    ],
    entry_points={
        'console_scripts': [
            'alpino-query = alpino_query.__main__:main'
        ]
    })



#!/usr/bin/env python3
import sys

from . import AlpinoQuery


def help():
    print("""
[action]
    mark
    subtree
    xpath

For mark:
    mark [inputxml] [tokens] [attributes]

For subtree:
    subtree [marked xml] [remove=rel/relcat/cat]

For xpath:
    xpath [subtree xml] [order=1/0]
""")


def main():
    if len(sys.argv) > 1:
        action = sys.argv[1]
    else:
        help()
        return

    query = AlpinoQuery()
    if action == "mark":
        [inputxml, tokens, attributes] = sys.argv[2:]
        query.mark(inputxml.replace('\\n', '\n'),
                   tokens.split(' '), attributes.split(' '))
        print(query.marked_xml)
    elif action == "subtree":
        [inputxml, remove] = sys.argv[2:]
        query.marked_xml = inputxml.replace('\\n', '\n')
        query.generate_subtree(remove)
        print(query.subtree_xml)
    elif action == "xpath":
        [inputxml, order] = sys.argv[2:]
        query.subtree_xml = inputxml.replace('\\n', '\n')
        query.generate_xpath(order in ['true', 'True', '1', 1, True])
        print(query.xpath)
    else:
        help()


if __name__ == "__main__":
    main()
